/* ====================================================================
 * Copyright (c) 2000 Jon Travis.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *        Jon Travis (jtravis@p00p.org) for use in the Apache 
 *        Project (http://www.apache.org)"
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "mod snake", "mod_snake", "modsnake" and "Jon Travis" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact Jon Travis (jtravis@p00p.org).
 *
 * 5. Products derived from this software may not be called "mod snake",
 *    "modsnake", or "mod_snake", nor may "mod snake", "modsnake" or
 *    "mod_snake" appear in their name, without prior written
 *    permission of the Jon Travis (jtravis@p00p.org).
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Jon Travis
 *     (jtravis@p00p.org) for use in the mod_snake module for
 *     the Apache Web Server (http://www.apache.org)"
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL JON TRAVIS OR HIS CONTRIBUTORS BE 
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, 
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, 
 * OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 */

#include "mod_snake.h"

/* This could be const, but Python doesn't dig that.  Python
   may modify this array of data in-place, similar to how Tcl_Eval does,
   so we may want to copy it before evaling it. -- XXX JMT 11-15-00 */
extern SnakePrecompiledPy Mod_Snake_Py;

/*
 * mod_snake_create_thread_and_lock:  Create a new thread state within an
 *                                    interpreter, set that thread current, and
 *                                    acquire the global interpreter lock.
 *
 * Arguments:                         interp = Interpreter within which to 
 *                                             create a new thread.
 *
 * Return values:                     Returns NULL on failure, else a valid
 *                                    threadstate on success.
 */

PyThreadState *mod_snake_create_thread_and_lock(PyInterpreterState *interp){
  PyThreadState *res;
  
  if((res = PyThreadState_New(interp)) == NULL)
    return NULL;

  PyEval_RestoreThread(res);
  return res;
}

/*
 * mod_snake_destroy_thread_and_unlock:  Destroy a thread state, and release
 *                                       the interpreter lock
 *
 * Arguments:                            tstate = Threadstate to destroy
 */

void mod_snake_destroy_thread_and_unlock(PyThreadState *tstate){
  PyThreadState_Clear(tstate);
  PyEval_SaveThread();
  PyThreadState_Delete(tstate);
}

/*
 * initmod_internal:  Load a statically compiled in module into the main
 *                    interp so that other modules have access to it.  
 *                    This sets up a dummy module and uses its dict to 
 *                    load all the proper info.  The Python global lock
 *                    MUST be acquired before calling this function.
 *
 * Arguments:         modname = Name to assign the module for 'import blah'
 *                    modstr  = Precompiled module information.
 *
 * Return values:  Returns -1 on failure, 0 on success.
 */

int mod_snake_initmod_internal(char *modname, SnakePrecompiledPy *modstr){
  PyObject *pmod, *pdict;

  if((pmod = mod_snake_exec_precompile(modname, modstr)) == NULL)
    return -1;
  
  pdict = PyModule_GetDict(pmod);   /* No Incref */
  PyDict_SetItemString(pdict, "__dummy__", Py_None);  /* Alert Python to dummy
                                                         module status */
  return 0;
}


/*
 * mod_snake_setup_syspath:  Setup the search path in the sys module.  If
 *                           any of the additions to the list fails, the
 *                           state of the list is undefined.  
 *
 * Arguments:                moddirs = array of module dirs to add to sys.path
 * 
 * Return values:            Returns -1 on failure, 0 on success.
 */

int mod_snake_setup_syspath(ap_array_header_t *moddirs){
  int i, err;
  PyObject *sysmod, *sysdict, *syspath, *str;

  /* Attempt to add search directories to sys.path.  The sys.path will
     be shared upon deletions and re-imports from other modules.  */
  
  if((sysmod = PyImport_ImportModule("sys")) == NULL)  /* Incref'd */
    return -1;

  if((sysdict = PyModule_GetDict(sysmod)) == NULL ||
     (syspath = PyDict_GetItemString(sysdict, "path")) == NULL){
    Py_DECREF(sysmod);
    return -1;
  }

  for(i=0; i< moddirs->nelts; i++){
    str = PyString_FromString(*((char **)moddirs->elts + i));
    err = PyList_Append(syspath, str);
    Py_DECREF(str);
    if(err){
      Py_DECREF(sysmod);
      return -1;
    }
  }

  Py_DECREF(sysmod);
  return 0;
}

/*
 * mod_snake_unload_module:  Unload a Python module by decref'ing all of the
 *                           objects it has stored.  The main thread state
 *                           for the module is destroyed, as well as the
 *                           interpreter.
 *
 * Arguments:                pymod = Module to unload
 *
 * Side Notes:               The Python interpreter lock MUST be held before
 *                           calling this function.
 */

void mod_snake_unload_module(ModSnakePyMod *pymod){
  PyThreadState *oldstate;
  int i;

  Py_DECREF(pymod->module);
  Py_DECREF(pymod->mod_obj);

  for(i=0; i< SNAKE_HOOK_NUM_HOOKS; i++)
    Py_XDECREF(pymod->hooks[i]);

  oldstate = PyThreadState_Swap(pymod->main_state);
  Py_EndInterpreter(pymod->main_state);
  PyThreadState_Swap(oldstate);
  
  pymod->valid                     = 0;
  pymod->module                    = NULL;
  pymod->mod_obj                   = NULL;
  pymod->interp                    = NULL;
  pymod->main_state                = NULL;

  for(i=0; i< SNAKE_HOOK_NUM_HOOKS; i++)
    pymod->hooks[i] = NULL;
}


/*
 * mod_snake_import_module:  Import a module into the server.  The module is 
 *                           imported into the current thread/interp.  On 
 *                           failure, pymod will be:
 *                           'the same as it ever was' (TM)
 *
 * Arguments:                 module = Module to load
 *                            s      = Server record to use for logging
 * 
 * Side Notes:                The Python interpreter lock MUST be held before
 *                            calling this function.
 *
 * Return values:             Returns -1 on failure, 0 on success.  
 */

int mod_snake_import_module(ModSnakePyMod *pymod, server_rec *s)
{
  PyObject *module, *reg_class, *py_modpy, *pargs, *mod_obj;
  char buf[1024], *cp;

  strncpy(buf, pymod->name, sizeof(buf) - 1);
  buf[sizeof(buf)-1] = '\0';
  if((cp = strchr(buf, '.')) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: module must be in the form of module.classname");
    return -1;
  }

  *cp = '\0';
  if((module = PyImport_ImportModule(buf)) == NULL){ /* Incref*/
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: error loading module '%s'", pymod->name);
    if(PyErr_Occurred()) PyErr_Print();
    return -1;      
  }

  /* Load the registration class */
  if((reg_class = PyObject_GetAttrString(module, cp + 1)) == NULL){ /* Incref*/
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: registration class '%s' not found in %s",
                 cp + 1, buf);
    Py_DECREF(module);
    if(PyErr_Occurred()) PyErr_Print();
    return -1;
  }

  /* Instantiate the class */
  py_modpy = mod_snake_get_modsnakepymod_shadowclass(pymod);
  pargs = Py_BuildValue("(O)", py_modpy); /* Incref */
  Py_DECREF(py_modpy);

  pymod->current_hook = SNAKE_HOOK_INIT;
  if((mod_obj = PyEval_CallObject(reg_class, pargs)) == NULL){ /* Incref */
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
                 "mod_snake: Error instantiating class '%s'", cp + 1);
    PyErr_Print();
    Py_DECREF(pargs);
    Py_DECREF(reg_class);
    Py_DECREF(module);
    pymod->current_hook = SNAKE_HOOK_NO_HOOK;
    return -1;
  }
  pymod->current_hook = SNAKE_HOOK_NO_HOOK;

  Py_DECREF(pargs);
  Py_DECREF(reg_class);
  
  pymod->mod_obj = mod_obj;
  pymod->module  = module;
  return 0;
}


/*
 * mod_snake_load_module:   Setup an interpreter, system path and
 *                          import internal modules into a server mod
 *                          that we're importing
 *
 * Arguments:               pymod   = Module to setup
 *                          syspath = System path to add
 *                          s       = Server for error reporting
 *                      
 *
 * Return values:           returns -1 on failure, 0 on success.
 */

int mod_snake_load_module(ModSnakePyMod *pymod, ap_array_header_t *syspath,
                          server_rec *s)
{
  PyThreadState *thr, *old_state;
  
  /* Setup a new interpreter for this module */
  if((thr = Py_NewInterpreter()) == NULL)
    return -1;
  
  /* Swap into the new interp so that all subsequent operations
     are performed within it */
  old_state = PyThreadState_Swap(thr);
  if(mod_snake_setup_syspath(syspath) == -1){
    Py_EndInterpreter(thr);
    PyThreadState_Swap(old_state);
    return -1;
  }
  
  initmod_snakec();

  if(mod_snake_initmod_internal("mod_snake", &Mod_Snake_Py) == -1)
  {
    Py_EndInterpreter(thr);
    PyThreadState_Swap(old_state);
    return -1;
  }

  if(mod_snake_import_module(pymod, s) == -1){
    Py_EndInterpreter(thr);
    PyThreadState_Swap(old_state);
    return -1;
  }


  pymod->interp     = thr->interp;
  pymod->main_state = thr;
  pymod->valid      = 1;
  PyThreadState_Swap(old_state);
  return 0;
}

static struct hook_mapping_st {
  const char *name;
  int hook_type;
} HookNameMappings[] = {
  { "create_dir_config", SNAKE_HOOK_BOFFO },/* SNAKE_HOOK_CREATE_DIR_CONFIG  */
  { "merge_dir_config",  SNAKE_HOOK_BOFFO },/* SNAKE_HOOK_MERGE_DIR_CONFIG   */
  { "open_logs",         SNAKE_HOOK_VOID  },/* SNAKE_HOOK_OPEN_LOGS          */
  { "header_parser",     SNAKE_HOOK_ALL   },/* SNAKE_HOOK_HEADER_PARSER      */
  { "post_read_request", SNAKE_HOOK_ALL   },/* SNAKE_HOOK_POST_READ_REQUEST  */
  { "log_transaction",   SNAKE_HOOK_ALL   },/* SNAKE_HOOK_LOG_TRANSACTION    */
  { "translate_name",    SNAKE_HOOK_FIRST },/* SNAKE_HOOK_TRANSLATE_NAME     */
  { "check_user_id",     SNAKE_HOOK_FIRST },/* SNAKE_HOOK_CHECK_USER_ID      */
  { "fixups",            SNAKE_HOOK_ALL   },/* SNAKE_HOOK_FIXUPS             */
  { "type_checker",      SNAKE_HOOK_FIRST },/* SNAKE_HOOK_TYPE_CHECKER       */
  { "access_checker",    SNAKE_HOOK_ALL   },/* SNAKE_HOOK_ACCESS_CHECKER     */
  { "auth_checker",      SNAKE_HOOK_FIRST },/* SNAKE_HOOK_AUTH_CHECKER       */
  { "content_handler",   SNAKE_HOOK_FIRST },/* SNAKE_HOOK_CONTENT_HANDLER    */
#ifndef APACHE13
  { "pre_connection",    SNAKE_HOOK_ALL   },/* SNAKE_HOOK_PRE_CONNECTION     */
  { "process_connection",SNAKE_HOOK_FIRST },/* SNAKE_HOOK_PROCESS_CONNECTION */
#endif
};


int mod_snake_hooknum_to_type(int hookno){
  if(hookno < 0 || hookno >= SNAKE_HOOK_NUM_HOOKS)
    return SNAKE_HOOK_BOFFO;
  
  return HookNameMappings[hookno].hook_type;
}
   
int mod_snake_hookname_to_num(const char *hookname){
  int i;

  for(i=0; i< SNAKE_HOOK_NUM_HOOKS; i++)
    if(!strcmp(hookname, HookNameMappings[i].name)) return i;

  return -1;
}

const char *mod_snake_hooknum_to_name(int hookno){
  if(hookno < 0 || hookno >= SNAKE_HOOK_NUM_HOOKS)
    return "Unknown hook type";

  return HookNameMappings[hookno].name;
}

PyObject * 
mod_snake_run_hook_create_dir_config(ModSnakePyMod *pymod, server_rec *s,
				     ModSnakeDirCfg *dir_cfg,
				     PyObject *callback)
{
  PyObject *pargs, *pres;
  PyThreadState *tstate;

  /* Lock */
  if((tstate = mod_snake_create_thread_and_lock(pymod->interp)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
		 "mod_snake: Failed to create a new threadstate");
    return NULL;
  }
  
  pargs = Py_BuildValue("(s)", dir_cfg->path);             /* pargs incref */
  pymod->current_hook = SNAKE_HOOK_CREATE_DIR_CONFIG;
  pres = PyEval_CallObject(callback, pargs);               /* Incref */
  pymod->current_hook = SNAKE_HOOK_NO_HOOK;
  Py_DECREF(pargs);
  if(pres == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
		 "mod_snake: (%s) Failed to call create_dir_config hook",
		 pymod->name);
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(tstate);
    return NULL;
  }
  mod_snake_destroy_thread_and_unlock(tstate);
  return pres;
}

PyObject * 
mod_snake_run_hook_merge_dir_config(ModSnakePyMod *pymod, server_rec *s,
				    PyObject *base_cdata, PyObject *new_cdata,
				    PyObject *callback)
{
  PyObject *pargs, *pres;
  PyThreadState *tstate;

  /* Lock */
  if((tstate = mod_snake_create_thread_and_lock(pymod->interp)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
		 "mod_snake: Failed to create a new threadstate");
    return NULL;
  }
  
  pargs = Py_BuildValue("(OO)", base_cdata, new_cdata);    /* pargs incref */
  pymod->current_hook = SNAKE_HOOK_MERGE_DIR_CONFIG;
  pres = PyEval_CallObject(callback, pargs);               /* Incref */
  pymod->current_hook = SNAKE_HOOK_NO_HOOK;
  Py_DECREF(pargs);
  if(pres == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, s,
		 "mod_snake: (%s) Failed to call create_dir_config hook",
		 pymod->name);
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(tstate);
    return NULL;
  }
  mod_snake_destroy_thread_and_unlock(tstate);
  return pres;
}

int
mod_snake_run_generic_conn_rec(ModSnakePyMod *pymod, conn_rec *c,
			       PyObject *callback, int cb_num)
{
  PyObject *pargs, *pres, *pconn_rec;
  PyThreadState *tstate;
  int res;

  /* Lock */
  if((tstate = mod_snake_create_thread_and_lock(pymod->interp)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, c->base_server,
		 "mod_snake: Failed to create a new threadstate");
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  
  if((pconn_rec = mod_snake_get_conn_rec_shadowclass(c)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, c->base_server,
		 "mod_snake: (%s) Error converting internal conn_rec rep",
		 pymod->name);
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(tstate);  /* Unlock */
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  
  pargs = Py_BuildValue("(O)", pconn_rec);  /* pargs + pconn_rec incref */
  Py_DECREF(pconn_rec);                   
  pymod->current_hook = cb_num;
  pres = PyEval_CallObject(callback, pargs);  /* Incref */
  pymod->current_hook = SNAKE_HOOK_NO_HOOK;
  Py_DECREF(pargs);
  if(pres == NULL || !PyInt_Check(pres)){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, c->base_server,
		 "mod_snake: (%s) Failed to call %s hook",
		 pymod->name, mod_snake_hooknum_to_name(cb_num));
    if(!PyErr_Occurred() && !PyInt_Check(pres)){
      PyErr_SetString(PyExc_TypeError, "Return value not an int");
      PyErr_Print();
    } else if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(tstate);   /* Unlock */
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  res = PyInt_AsLong(pres);
  mod_snake_destroy_thread_and_unlock(tstate);     /* Unlock */
  return res;
}


int mod_snake_run_generic_request_rec(ModSnakePyMod *pymod, request_rec *r,
				      ModSnakeDirCfg *per_dir_cfg,
				      PyObject *callback, int cb_num)
{
  ap_array_header_t *dir_cfgs;
  PyObject *pargs, *pres, *prequest_rec, *per_dir_data;
  PyThreadState *tstate;
  int res, i;

  /* Lock */
  if((tstate = mod_snake_create_thread_and_lock(pymod->interp)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, r->server,
		 "mod_snake: Failed to create a new threadstate");
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  
  if((prequest_rec = mod_snake_get_request_rec_shadowclass(r)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, r->server,
		 "mod_snake: (%s) Error converting internal request_rec rep",
		 pymod->name);
    if(PyErr_Occurred()) PyErr_Print();
    mod_snake_destroy_thread_and_unlock(tstate);  /* Unlock */
    return HTTP_INTERNAL_SERVER_ERROR;
  }

  /* Locate the per-dir callback data for this module */
  per_dir_data = Py_None;
  dir_cfgs = per_dir_cfg->dir_config_data;
  for(i=0; i< dir_cfgs->nelts; i++){
    ModSnakeCallbackData *cbdata = (ModSnakeCallbackData *)dir_cfgs->elts + i;

    if(cbdata->module == pymod){
      per_dir_data = cbdata->data;
      break;
    }
  }
  
  pargs = Py_BuildValue("(OO)", per_dir_data, prequest_rec); 
  /* pargs + pcb_data + prequest_rec incref */
  Py_DECREF(prequest_rec);                   
  pymod->current_hook = cb_num;
  pres = PyEval_CallObject(callback, pargs);  /* Incref */
  pymod->current_hook = SNAKE_HOOK_NO_HOOK;
  Py_DECREF(pargs);
  if(pres == NULL || !PyInt_Check(pres)){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 0, r->server,
		 "mod_snake: (%s) Failed to call %s hook",
		 pymod->name, mod_snake_hooknum_to_name(cb_num));
    if(!PyErr_Occurred() && !PyInt_Check(pres)){
      PyErr_SetString(PyExc_TypeError, "Return value not an int");
      PyErr_Print();
    } else if(PyErr_Occurred()) 
      PyErr_Print();
    mod_snake_destroy_thread_and_unlock(tstate);   /* Unlock */
    return HTTP_INTERNAL_SERVER_ERROR;
  }
  res = PyInt_AsLong(pres);
  mod_snake_destroy_thread_and_unlock(tstate);     /* Unlock */
  return res;
}

void mod_snake_run_hook_open_logs(ModSnakePyMod *pymod)
{
  PyObject *pargs, *pres, *ppymod;
  PyThreadState *tstate;

  /* Lock */
  if((tstate = mod_snake_create_thread_and_lock(pymod->interp)) == NULL){
    mod_snake_log_error(APLOG_MARK, APLOG_NOERRNO|APLOG_CRIT, 
		 0, pymod->parentsvr->server,
		 "mod_snake: Failed to create a new threadstate");
    return;
  }
  
  ppymod = mod_snake_get_modsnakepymod_shadowclass(pymod);
  pargs = Py_BuildValue("(O)", ppymod); /* pargs + ppymod incref */
  Py_DECREF(ppymod);
  pymod->current_hook = SNAKE_HOOK_OPEN_LOGS;
  pres = PyEval_CallObject(pymod->hooks[SNAKE_HOOK_OPEN_LOGS], 
			   pargs); /* INCREF */
  pymod->current_hook = SNAKE_HOOK_NO_HOOK;
  Py_DECREF(pargs);
  if(pres == NULL){
    if(PyErr_Occurred())
      PyErr_Print();
  } else
    Py_DECREF(pres);  /* Discard the result */
  mod_snake_destroy_thread_and_unlock(tstate);     /* Unlock */
  return;
}

